#include <QMessageBox>
#include <math.h>

#include "geommodel.h"
#include "compgeom.h"

#include "curves/line.h"

#ifndef ABS
#define ABS(x) (((x) < 0)? -(x): (x))
#endif

#define ABSTOL (1e-7)


//--------------------------------------------------------------------
GeomModel::GeomModel()
{
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
GeomModel::~GeomModel()
{
 delAllCurves();
 delAllRegions();
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GeomModel::insertCurve( Curve* _curve )
{
#if 0
 // Currently only line segments are treated.
 if( _curve->getType() != Curve::LINE )
  return;

 bool isCollinear = false;

 // Get given curve points.
 vector<Point> pts1 = _curve->getPoints();

 // Check for collinear elements.
 for(int i = 0; i < m_curves.size(); i++)
 {
  if( m_curves[i]->getType() == Curve::LINE )
  {
   // Get line points.
   vector<Point> pts2 = m_curves[i]->getPoints();
   isCollinear = CompGeom::checkCollinearSegments(
                                pts1[0], pts1[1], pts2[0], pts2[1] );
   if( isCollinear )
    break;
  }
 }
 if( !isCollinear )
  m_curves.push_back(_curve);
#else
 m_curves.push_back(_curve);
#endif
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GeomModel::delCurve( int curve_id )
{
 delete m_curves[curve_id];
 m_curves.erase( m_curves.begin()+curve_id );

 // Update curve information of regions in response to curve deletion.
 // In case a region uses the deleted curve, also delete this region.
 // Otherwise, update curve indexes of region.
 bool regionRemoved;
 do
 {
  regionRemoved = false;
  for(int i = 0; i < m_regions.size(); i++)
  {
   vector<int> curveIds = m_regions[i]->getCurveIds();
   for(int j = 0; j < curveIds.size(); j++)
   {
    if(curveIds[j] == curve_id)
    {
     delRegion(i);
     regionRemoved = true;
     break;
    }
   }
   if( regionRemoved )
   {
    break;
   }
   else
   {
    for(int j = 0; j < curveIds.size(); j++)
    {
     if(curveIds[j] > curve_id)
      curveIds[j] -= 1;
    }
    m_regions[i]->setCurveIds(curveIds);
   }
  }
 } while( regionRemoved );

}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GeomModel::delAllCurves()
{
 while( !m_curves.empty() )
 {
  delete m_curves.back();
 }
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GeomModel::delSelectedCurves()
{
 bool curveRemoved;
 do
 {
  curveRemoved = false;
  for(int i = 0; i < m_curves.size(); i++)
  {
   if( m_curves[i]->isSelected() )
   {
    delCurve(i);
    curveRemoved = true;
    break;
   }
  }
 } while( curveRemoved );
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GeomModel::delRegion(int region_id)
{
 delete m_regions[region_id];
 m_regions.erase(m_regions.begin()+region_id);
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GeomModel::delAllRegions()
{
 while( !m_regions.empty() )
 {
  delete m_regions.back();
 }
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GeomModel::delSelectedRegions()
{
 bool regionRemoved;
 do
 {
  regionRemoved = false;
  for(int i = 0; i < m_regions.size(); i++)
  {
   if( m_regions[i]->isSelected() )
   {
    delRegion(i);
    regionRemoved = true;
    break;
   }
  }
 } while( regionRemoved );
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GeomModel::getBoundBox(double* _xmin, double* _xmax, double* _ymin, double* _ymax)
{
 if( (m_curves.size() < 1) && (m_regions.size() < 1) )
 {
  *_xmin =  0.0;
  *_xmax = 10.0;
  *_ymin =  0.0;
  *_ymax = 10.0;
  return;
 }

 if( m_curves.size() > 0 )
 {
  m_curves[0]->getBoundBox(_xmin, _xmax, _ymin, _ymax);
  for(int i = 1; i < m_curves.size(); i++)
  {
   double xmin_c, xmax_c, ymin_c, ymax_c;
   m_curves[i]->getBoundBox(&xmin_c, &xmax_c, &ymin_c, &ymax_c);
   if(xmin_c < *_xmin)
    *_xmin = xmin_c;
   if(xmax_c > *_xmax)
    *_xmax = xmax_c;
   if(ymin_c < *_ymin)
    *_ymin = ymin_c;
   if(ymax_c > *_ymax)
    *_ymax = ymax_c;
  }
 }

 if( m_regions.size() > 0 )
 {
  int i_init;
  if( m_curves.size() == 0 )
  {
   m_regions[0]->getBoundBox(_xmin, _xmax, _ymin, _ymax);
   i_init = 1;
  }
  else
  {
   i_init = 0;
  }
  for(int i = i_init; i < m_regions.size(); i++)
  {
   double xmin_r, xmax_r, ymin_r, ymax_r;
   m_regions[i]->getBoundBox(&xmin_r, &xmax_r, &ymin_r, &ymax_r);
   if(xmin_r < *_xmin)
    *_xmin = xmin_r;
   if(xmax_r > *_xmax)
    *_xmax = xmax_r;
   if(ymin_r < *_ymin)
    *_ymin = ymin_r;
   if(ymax_r > *_ymax)
    *_ymax = ymax_r;
  }
 }
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GeomModel::selectPick( double _x, double _y, double _tol, bool _shiftkey )
{
 if( (m_curves.size() < 1) && (m_regions.size() < 1) )
  return;

 double xC, yC;
 int id_target = -1;
 double dmin = _tol;
 for(int i = 0; i < m_curves.size(); i++)
 {
  // Compute distance between given point and curve and
  // update minimum distance
  xC = _x;
  yC = _y;
  double d = m_curves[i]->closestPoint( &xC, &yC );
  if(d < dmin)
  {
   dmin = d;
   id_target = i;
  }
 }

 // Revert selection of picked curve
 if(id_target > -1)
 {
  if( m_curves[id_target]->isSelected() )
  {
   m_curves[id_target]->setSelected(false);
  } 
  else
  {
   m_curves[id_target]->setSelected(true);
  }
 }

 if( !_shiftkey )
 {
  // If shift key is not pressed, unselect all curves except 
  // the picked one (if there was one selected)
  for(int i = 0; i < m_curves.size(); i++)
  {
   if( i != id_target )
    m_curves[i]->setSelected(false);
  }
 }

 // Check whether point is inside a region.
 Point p(_x,_y);
 for(int i = 0; i < m_regions.size(); i++)
 {
  vector<Point> pts = m_regions[i]->getPoints();
  if( CompGeom::isPointInPolygon( pts, p ) )
  {
   if( m_regions[i]->isSelected() )
    m_regions[i]->setSelected(false);
   else
    m_regions[i]->setSelected(true);
  }
  else
  {
   if( !_shiftkey )
    m_regions[i]->setSelected(false);
  }
 }

}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GeomModel::selectFence( double _xmin, double _xmax,
                        double _ymin, double _ymax, bool _shiftkey )
{
 if( (m_curves.size() < 1) && (m_regions.size() < 1) )
  return;

 bool inFence;
 double xmin_c, xmax_c, ymin_c, ymax_c;
 for(int i = 0; i < m_curves.size(); i++)
 {
  m_curves[i]->getBoundBox(&xmin_c, &xmax_c, &ymin_c, &ymax_c);
  if( (xmin_c < _xmin) || (xmax_c > _xmax) ||
      (ymin_c < _ymin) || (ymax_c > _ymax) )
   inFence = false;
  else
   inFence = true;

  if( inFence )
  {
   // Select curve inside fence
   m_curves[i]->setSelected(true);
  }
  else
  {
  // If shift key is not pressed, unselect curve outside fence
   if( !_shiftkey )
    m_curves[i]->setSelected(false);
  }
 }

 double xmin_r, xmax_r, ymin_r, ymax_r;
 for(int i = 0; i < m_regions.size(); i++)
 {
  m_regions[i]->getBoundBox(&xmin_r, &xmax_r, &ymin_r, &ymax_r);
  if( (xmin_r < _xmin) || (xmax_r > _xmax) ||
      (ymin_r < _ymin) || (ymax_r > _ymax) )
   inFence = false;
  else
   inFence = true;

  if( inFence )
  {
   // Select region inside fence
   m_regions[i]->setSelected(true);
  }
  else
  {
  // If shift key is not pressed, unselect region outside fence
   if( !_shiftkey )
    m_regions[i]->setSelected(false);
  }
 }

}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
bool 
GeomModel::snapToCurve( double* _x, double* _y, double _tol )
{
 if(m_curves.size() < 1)
  return false;

 double xC, yC;
 double xClst = *_x;
 double yClst = *_y;
 int id_target = -1;
 double dmin = _tol;
 double d;

 for(int i = 0; i < m_curves.size(); i++)
 {
  // Check to see whether point is close (within given tolerance)
  // to one of curve vertices. If that is the case, continue to 
  // the next curve.
  xC = m_curves[i]->getXinit();
  yC = m_curves[i]->getYinit();
  if( (ABS(*_x-xC) < _tol) && (ABS(*_y-yC) < _tol) )
  {
   d = sqrt((*_x-xC)*(*_x-xC)+(*_y-yC)*(*_y-yC));
   if(d < dmin)
   {
    xClst = xC;
    yClst = yC;
    dmin = d;
    id_target = i;
   }
   continue;
  }

  xC = m_curves[i]->getXend();
  yC = m_curves[i]->getYend();
  if( (ABS(*_x-xC) < _tol) && (ABS(*_y-yC) < _tol) )
  {
   d = sqrt((*_x-xC)*(*_x-xC)+(*_y-yC)*(*_y-yC));
   if(d < dmin)
   {
    xClst = xC;
    yClst = yC;
    dmin = d;
    id_target = i;
   }
   continue;
  }

  // Compute distance between given point and curve and
  // update minimum distance
  xC = *_x;
  yC = *_y;
  d = m_curves[i]->closestPoint( &xC, &yC );
  if(d < dmin)
  {
   xClst = xC;
   yClst = yC;
   dmin = d;
   id_target = i;
  }
 }

 if( id_target < 0 )
  return false;

 // If found a closest point on a curve, return its 
 // coordinates in the given variables.
 *_x = xClst;
 *_y = yClst;
 return true;
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
bool 
GeomModel::intersectTwoSelectedLines()
{
 if(m_curves.size() < 1)
  return false;

 // Check for exactly two selected lines.
 int i_target1 = -1;
 int i_target2 = -1;
 for(int i = 0; i < m_curves.size(); i++)
 {
  if( m_curves[i]->isSelected() )
  {
   if( i_target1 == -1 )
   {
    i_target1 = i;
   }
   else if ( i_target2 == -1 )
   {
    i_target2 = i;
   }
   else  // more then two selected lines
   {
    QMessageBox msgBox;
    msgBox.setText("Cannot perform curve intersection:\n"
                   "More than two curves were selected.");
    msgBox.exec();
    return false;
   }
  }
 }

 // Check for two selected lines.
 if( i_target2 == -1 )
 {
  QMessageBox msgBox;
  msgBox.setText("Cannot perform curve intersection:\n"
                 "Exactly two curves should be selected.");
  msgBox.exec();
  return false;
 }

 // Currently only intersection of line segments are treated.
 if( (m_curves[i_target1]->getType() != Curve::LINE) ||
     (m_curves[i_target2]->getType() != Curve::LINE) )
 {
  QMessageBox msgBox;
  msgBox.setText("Cannot perform curve intersection:\n"
                 "Currently only intersection of line segments are treated.");
  msgBox.exec();
  return false;
 }

 // Get lines' points.
 vector<Point> pts1 = m_curves[i_target1]->getPoints();
 vector<Point> pts2 = m_curves[i_target2]->getPoints();

 // Compute intersection between two lines.
 Point pi;
 double t1, t2;
 int status = CompGeom::computeSegmentSegmentIntersection(
                         pts1[0], pts1[1], pts2[0], pts2[1], pi, t1, t2 );
 Line* lnA;
 Line* lnB;
 Line* lnC;
 Line* lnD;
 switch( status )
 {
  default:
  case CompGeom::DO_NOT_INTERSECT:
  case CompGeom::COLLINEAR:
  case CompGeom::PARALLEL:
   {
    QMessageBox msgBox;
    msgBox.setText("Cannot perform line intersection:\n"
                   "The two selected lines do not intercept.");
    msgBox.exec();
    return false;
   }
   break;
  case CompGeom::DO_INTERSECT:
   {
    lnA = new Line( pts1[0].getX(), pts1[0].getY(), pi.getX(), pi.getY() );
    lnB = new Line( pi.getX(), pi.getY(), pts1[1].getX(), pts1[1].getY() );
    lnC = new Line( pts2[0].getX(), pts2[0].getY(), pi.getX(), pi.getY() );
    lnD = new Line( pi.getX(), pi.getY(), pts2[1].getX(), pts2[1].getY() );
    delCurve(i_target1);
    i_target2 = -1;
    for(int i = 0; i < m_curves.size(); i++)
    {
     if( m_curves[i]->isSelected() )
      i_target2 = i;
    }
    if( i_target2 != -1 )
     delCurve(i_target2);
    m_curves.push_back(lnA);
    m_curves.push_back(lnB);
    m_curves.push_back(lnC);
    m_curves.push_back(lnD);
    return true;
   }
   break;
  case CompGeom::TOUCH:
   {
    bool createdLines = false;
    if( (t1 > 0.0) && (t1 < 1.0) )
    {
     lnA = new Line( pts1[0].getX(), pts1[0].getY(), pi.getX(), pi.getY() );
     lnB = new Line( pi.getX(), pi.getY(), pts1[1].getX(), pts1[1].getY() );
     delCurve(i_target1);
     m_curves.push_back(lnA);
     m_curves.push_back(lnB);
     createdLines = true;
    }
    if( (t2 > 0.0) && (t2 < 1.0) )
    {
     lnC = new Line( pts2[0].getX(), pts2[0].getY(), pi.getX(), pi.getY() );
     lnD = new Line( pi.getX(), pi.getY(), pts2[1].getX(), pts2[1].getY() );
     delCurve(i_target2);
     m_curves.push_back(lnC);
     m_curves.push_back(lnD);
     createdLines = true;
    }
    if( createdLines )
     return true;
   }
   break;
 }
 return false;
}
//--------------------------------------------------------------------

//--------------------------------------------------------------------
void 
GeomModel::createRegion()
{
    vector <int> curveRegion;
    vector <bool> curveOrientation;
    list <int> selInd;
    int masterID = -1;
    int num_sel_curves;

    // Get all selected curves.
    for (int  i = 0; i<m_curves.size();i++)
    {
       if (m_curves[i]->isSelected())
       {
         if (masterID == -1)
         {
           // The master curve of region is the first curve found selected.
           masterID = i;
         }
         else
         {
           // Save all selected curves, except the master one in a list.
           selInd.push_back(i);
         }
       }
    }

    // Check for no curve selected.
    if (masterID == -1)
    {
     QMessageBox msgBox;
     msgBox.setText("Cannot create region:\n"
                    "No curve is selected.");
     msgBox.exec();
     return;
    }

    // Get number of selected lines.
    num_sel_curves = selInd.size() + 1;

    // Put the index of the master curve in a list of region curves.
    // First region point is the first point of the master curve.
    curveRegion.push_back(masterID);
    curveOrientation.push_back(true);
    double fst_x = m_curves[masterID]->getXinit();
    double fst_y = m_curves[masterID]->getYinit();
    double cur_x = m_curves[masterID]->getXend();
    double cur_y = m_curves[masterID]->getYend();
    // Check for a single closed curve, in which case create region.
    if( (fst_x == cur_x) && (fst_y == cur_y) )
    {
      // If closed curve is the only select, create a region.
      // Otherwise, abort region creation.
      if( num_sel_curves == 1 )
      {
        vector<Point> regionPts = m_curves[masterID]->getPointsToDraw();
        vector<int> regionCurveIds;
        regionCurveIds.push_back(masterID);
        vector<bool> regionCurveOrients;
        // assume that closed curve points a in counter-clockwise order
        regionCurveOrients.push_back(true);
        Region* region = new Region( regionPts,
                                     regionCurveIds, regionCurveOrients );
        m_regions.push_back(region);
      }
      else
      {
       QMessageBox msgBox;
       msgBox.setText("Cannot create region:\n"
                      "Select only one closed curve.");
       msgBox.exec();
      }
      return;
    }
    bool found_next_curve = false;
    double orientedArea = 0.0; // oriented area of polygon

    // Traverse list of selected curves trying to form a continuous chain.
    do
    {
      double prv_x = cur_x;
      double prv_y = cur_y;
      found_next_curve = false;
      for (list <int>::iterator it = selInd.begin(); it != selInd.end(); it++)
      {
        double x1 = m_curves[*it]->getXinit();
        double y1 = m_curves[*it]->getYinit();
        double x2 = m_curves[*it]->getXend();
        double y2 = m_curves[*it]->getYend();
        // Check for closed curve, in which case abort region creation.
        if( (x1 == x2) && (y1 == y2) )
        {
         QMessageBox msgBox;
         msgBox.setText("Cannot create region:\n"
                        "Select only one closed curve.");
         msgBox.exec();
         return;
        }
        bool curve_orientation = true;
        if ((ABS(cur_x-x1) < ABSTOL) && (ABS(cur_y-y1) < ABSTOL))
        {
          cur_x = x2;
          cur_y = y2;
          curve_orientation = true;
          found_next_curve = true;
        }
        else if ((ABS(cur_x-x2) < ABSTOL) && (ABS(cur_y-y2) < ABSTOL))
        {
          cur_x = x1;
          cur_y = y1;
          curve_orientation = false;
          found_next_curve = true;
        }
        // If found curve along chain of curves, insert it in region and
        // remove it from list of selected curves.
        if (found_next_curve)
        {
          curveRegion.push_back(*it);
          it = selInd.erase(it);
          curveOrientation.push_back(curve_orientation);

          // Also update the oriented area of polygon.
          orientedArea += (prv_x-fst_x)*(cur_y-fst_y) -
                          (cur_x-fst_x)*(prv_y-fst_y);
          break;
        }
      }
    } while (found_next_curve);

    // Check whether number of selected curves is the same as the 
    // number of region curves.
    // Essentially, this is checking whether all selected curves form
    // a continuous chain.
    if (num_sel_curves != curveRegion.size())
    {
      QMessageBox msgBox;
      msgBox.setText("Cannot create region:\n"
                     "The selected curves do not form a continuous chain.");
      msgBox.exec();
      return;
    }

    // Check whether first point is equal to last curve point.
    if ((ABS(fst_x-cur_x) < ABSTOL) && (ABS(fst_y-cur_y) < ABSTOL))
    {
      // The selected curves form a polygon: a continous and closed chain.
      // Always store region points in counter-closewise order.
      // Check whether polygon points are in counter-clockwise order.
      // This is known using the sign of the computed oriented area.
      // If it is positive, points are in counter-clockwise order.
      // In this case, store region points in the same order of the
      // traversed selected curves.
      // Othewise, store region points in the reverse order of the
      // traversed selected curves.
      vector<Point> regionPts;
      vector<int> regionCurveIds;
      vector<bool> regionCurveOrients;
      if( orientedArea > 0.0 )
      {
        for (int i = 0; i < curveRegion.size(); i++)
        {
          regionCurveIds.push_back(curveRegion[i]);
          vector<Point> curvePts = m_curves[curveRegion[i]]->getPointsToDraw();
          if (curveOrientation[i])
          {
            for (int j = 0; j < curvePts.size()-1; j++)
            {
              regionPts.push_back(curvePts[j]);
            }
            regionCurveOrients.push_back(true);
          } 
          else
          {
            for (int j = curvePts.size()-1; j > 0 ; j--)
            {
              regionPts.push_back(curvePts[j]);
            }
            regionCurveOrients.push_back(false);
          }
        }
      }
      else
      {
        for (int i = curveRegion.size()-1; i >= 0 ; i--)
        {
          regionCurveIds.push_back(curveRegion[i]);
          vector<Point> curvePts = m_curves[curveRegion[i]]->getPointsToDraw();
          if (curveOrientation[i])
          {
            for (int j = curvePts.size()-1; j > 0 ; j--)
            {
              regionPts.push_back(curvePts[j]);
            }
            regionCurveOrients.push_back(false);
          } 
          else
          {
            for (int j = 0; j < curvePts.size()-1; j++)
            {
              regionPts.push_back(curvePts[j]);
            }
            regionCurveOrients.push_back(true);
          }
        }
      }
      Region* region = new Region( regionPts,
                                   regionCurveIds, regionCurveOrients );
      m_regions.push_back(region);
    }
    else
    {
      QMessageBox msgBox;
      msgBox.setText("Cannot create region:\n"
                     "The selected curves do not close off a region.");
      msgBox.exec();
      return;
    }
}
